from liblaf.grapes import deps

with deps.optional_imports(extra="duration"):
    import about_time


def human_count(count: int, unit: str = "", prec: int | None = None) -> str:
    """Convert a numerical count into a human-readable string.

    Args:
        count: The numerical count to be converted.
        unit: The unit of measurement to be appended to the count.
        prec: The precision for the human-readable format. If None, default precision is used.

    Returns:
        The human-readable string representation of the count.
    """
    # TODO: remove dependency on `about-time`
    hc = about_time.HumanCount(count, unit)
    return hc.as_human(prec)



import inspect
from collections.abc import Callable, Mapping, Sequence
from pathlib import Path
from typing import Any, Literal

import autoregistry
import rich
import rich.highlighter
from rich.style import Style
from rich.text import Text

_func = autoregistry.Registry(prefix="_func_")


def func(obj: Callable, *, style: Literal["short", "long"] = "short") -> Text:
    return _func[style](obj)


@_func
def _func_short(obj: Callable) -> Text:
    obj = inspect.unwrap(obj)
    text = Text()
    name: str = _get_name(obj)
    source_file: Path | None = _get_source_file(obj)
    lineno: int = _get_source_lineno(obj)
    if source_file and lineno:
        text.append(
            f"{name}()",
            style=Style(link=f"{source_file.as_uri()}#{lineno}"),
        )
    else:
        text.append(f"{name}()")
    return text


@_func
def _func_long(obj: Callable) -> Text:
    obj = inspect.unwrap(obj)
    text = Text()
    module: str = _get_module(obj)
    qualname: str = _get_qualname(obj)
    source_file: Path | None = _get_source_file(obj)
    lineno: int = _get_source_lineno(obj)
    if source_file:
        text.append(module, style=Style(link=source_file.as_uri()))
        text.append(".")
        text.append(
            f"{qualname}(...)",
            style=Style(link=f"{source_file.as_uri()}#{lineno}"),
        )
    else:
        text.append(f"{module}.{qualname}(...)")
    return text


highlighter = rich.highlighter.ReprHighlighter()


def call(fn: Callable, args: Sequence, kwargs: Mapping) -> Text:
    args, kwargs = _bind_safe(fn, args, kwargs)
    fn = inspect.unwrap(fn)
    name: str = _get_name(fn)
    source_file: Path | None = _get_source_file(fn)
    lineno: int = _get_source_lineno(fn)
    text: Text = Text()
    if source_file:
        text.append(name, style=Style(link=f"{source_file.as_uri()}#{lineno}"))
    else:
        text.append(name)
    text.append("(")
    parts: list[str] = [repr(value) for value in args]
    parts += [f"{key}={value!r}" for key, value in kwargs.items()]
    text.append(", ".join(parts))
    text.append(")")
    text = highlighter(text)
    return text


def _bind_safe(
    fn: Callable, args: Sequence, kwargs: Mapping
) -> tuple[Sequence, Mapping]:
    try:
        signature: inspect.Signature = inspect.signature(fn)
        arguments: inspect.BoundArguments = signature.bind(*args, **kwargs)
    except TypeError:
        return args, kwargs
    else:
        return arguments.args, arguments.kwargs


def _get_module(obj: Any) -> str:
    return getattr(obj, "__module__", "unknown")


def _get_name(obj: Any) -> str:
    return getattr(obj, "__name__", "<unknown>")


def _get_qualname(obj: Any) -> str:
    return getattr(obj, "__qualname__", "<unknown>")


def _get_source_file(obj: Any) -> Path | None:
    try:
        if source_file := inspect.getsourcefile(obj):
            return Path(source_file)
    except TypeError:
        pass
    return None


def _get_source_lineno(obj: Any) -> int:
    try:
        lineno: int
        _lines, lineno = inspect.getsourcelines(obj)
    except (OSError, TypeError):
        pass
    else:
        return lineno
    return 0



import functools
from collections.abc import Callable

from liblaf.grapes import pretty
from liblaf.grapes.logging import depth_tracker

from ._base import BaseTimer


class TimedFunction[**P, T]:
    timing: BaseTimer
    __wrapped__: Callable[P, T]

    def __init__(self, fn: Callable[P, T], /, *, timing: BaseTimer) -> None:
        self.timing = timing
        if self.timing.label is None:
            self.timing.label = pretty.func(fn).plain or "Function"
        functools.update_wrapper(self, fn)

    @depth_tracker
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        self.timing.start()
        result: T = self.__wrapped__(*args, **kwargs)
        self.timing.stop()
        return result



from collections.abc import Iterable
from typing import Any

from liblaf.grapes.typed import ClassInfo


def as_iterable(obj: Any, base_type: ClassInfo | None = (str, bytes)) -> Iterable:
    # https://more-itertools.readthedocs.io/en/stable/api.html#more_itertools.always_iterable
    if obj is None:
        return ()
    if base_type is not None and isinstance(obj, base_type):
        return (obj,)
    if isinstance(obj, Iterable):
        return obj
    return (obj,)



import os
import time

import autoregistry

type TimerName = str


_get_time = autoregistry.Registry()
_get_time["children_system"] = lambda: os.times().children_system
_get_time["children_user"] = lambda: os.times().children_user
_get_time["elapsed"] = lambda: os.times().elapsed
_get_time["monotonic"] = time.monotonic
_get_time["perf"] = time.perf_counter
_get_time["process"] = time.process_time
_get_time["system"] = lambda: os.times().system
_get_time["thread"] = time.thread_time
_get_time["time"] = time.time
_get_time["user"] = lambda: os.times().user


def get_time(timer: TimerName = "perf") -> float:
    return _get_time[timer]()



import functools
import sys
from typing import IO, Literal, TypedDict, Unpack

import rich
from environs import env
from rich.console import Console
from rich.style import Style
from rich.theme import Theme

from liblaf.grapes import path
from liblaf.grapes.typed import PathLike


def theme() -> Theme:
    return Theme(
        {
            "logging.level.notset": Style(dim=True),
            "logging.level.trace": Style(color="cyan", bold=True),
            "logging.level.debug": Style(color="blue", bold=True),
            "logging.level.icecream": Style(color="magenta", bold=True),
            "logging.level.info": Style(bold=True),
            "logging.level.success": Style(color="green", bold=True),
            "logging.level.warning": Style(color="yellow", bold=True),
            "logging.level.error": Style(color="red", bold=True),
            "logging.level.critical": Style(color="red", bold=True, reverse=True),
        },
        inherit=True,
    )


class ConsoleKwargs(TypedDict, total=False):
    theme: Theme | None
    stderr: bool | None
    file: IO | PathLike | None
    width: int | None


@functools.cache
def get_console(**kwargs: Unpack[ConsoleKwargs]) -> Console:
    kwargs.setdefault("theme", theme())
    file: IO | PathLike | None = kwargs.get("file")
    if _is_stdout(**kwargs):
        kwargs = _ci(**kwargs)
        rich.reconfigure(**kwargs)
        return rich.get_console()
    if _is_stderr(**kwargs):
        kwargs = _ci(**kwargs)
        kwargs.pop("file", None)
        kwargs["stderr"] = True
        return Console(**kwargs)  # pyright: ignore[reportArgumentType]
    if path.is_path_like(file):
        kwargs["file"] = path.as_path(file).open("w")
    return Console(**kwargs)  # pyright: ignore[reportArgumentType]


def _is_stdout(**kwargs: Unpack[ConsoleKwargs]) -> bool:
    file: IO | PathLike | None = kwargs.get("file")
    return (
        (file is None and not kwargs.get("stderr", False))
        or (file in ("stdout", "<stdout>", sys.stdout))
        or getattr(file, "name", None) in ("stdout", "<stdout>")
    )


def _is_stderr(**kwargs: Unpack[ConsoleKwargs]) -> bool:
    file: IO | PathLike | None = kwargs.get("file")
    return (
        (file is None and kwargs.get("stderr", False))
        or (file in ("stderr", "<stderr>", sys.stderr))
        or getattr(file, "name", None) in ("stderr", "<stderr>")
    )


def _ci(**kwargs: Unpack[ConsoleKwargs]) -> ConsoleKwargs:
    if not env.bool("GITHUB_ACTIONS", False):
        return kwargs
    kwargs.setdefault("width", 128)
    return kwargs


def force_terminal(file: Literal["stdout", "stderr"] | IO | PathLike) -> bool | None:
    """...

    References:
        1. <https://force-color.org/>
        2. <https://no-color.org/>
    """
    if file not in ("stdout", "stderr"):
        return None
    if env.bool("FORCE_COLOR", None):
        return True
    if env.bool("NO_COLOR", None):
        return False
    if env.bool("GITHUB_ACTIONS", None):
        return True
    return None



from pathlib import Path
from types import FrameType
from typing import Literal

from loguru._get_frame import get_frame
from rich.style import Style
from rich.text import Text

from liblaf.grapes.typed import PathLike


def location(
    name: str | None,
    function: str | None,
    line: int | None,
    file: PathLike | None = None,
    *,
    enable_link: bool = True,
    style: Literal["long", "short"] = "long",
) -> Text:
    text = Text()
    file: Path | None = Path(file) if file is not None else None
    function = function or "<unknown>"
    line = line or 0
    if (style == "short") and (name is not None):
        name = name.split(".")[-1]
    if enable_link and file is not None and file.exists():
        text.append(
            f"{name}:{function}:{line}", style=Style(link=f"{file.as_uri()}#{line}")
        )
    else:
        text.append(f"{name}:{function}:{line}")
    return text


def caller_location(depth: int = 1, style: Literal["long", "short"] = "long") -> Text:
    frame: FrameType | None
    try:
        frame = get_frame(depth)
    except ValueError:
        frame = None
    file: str | None = None
    function: str | None = None
    line: int | None = None
    name: str | None = None
    if frame is not None:
        file = frame.f_code.co_filename
        function = frame.f_code.co_name
        line = frame.f_lineno
        name = frame.f_globals.get("__name__", "unknown")
    text: Text = location(
        function=function, line=line, name=name, file=file, style=style
    )
    return text



import functools
import importlib.util


@functools.lru_cache
def has_module(name: str, package: str | None = None) -> bool:
    """Check if a module can be imported.

    Args:
        name: The name of the module to check.
        package: The package name to use as the anchor point from which to resolve the relative module name. Defaults to `None`.

    Returns:
        `True` if the module can be imported, `False` otherwise.
    """
    return importlib.util.find_spec(name, package) is not None



import contextlib
import functools
import types
from collections.abc import Callable, Iterable
from typing import Any, Self, override

import attrs

from liblaf.grapes.error import DispatchLookupError
from liblaf.grapes.logging import depth_tracker

from ._base import BaseTimer
from ._function import TimedFunction
from ._iterable import TimedIterable


@attrs.define
class Timer(contextlib.AbstractContextManager, BaseTimer):
    @override
    @depth_tracker
    def __enter__(self) -> Self:
        self.start()
        return self

    @override
    @depth_tracker
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
        /,
    ) -> None:
        self.stop()

    @functools.singledispatchmethod
    def __call__(self, *args, **kwargs) -> Any:
        raise DispatchLookupError(self.__call__, args, kwargs)

    @__call__.register(Callable)
    def _[**P, T](self, func: Callable[P, T], /, **kwargs) -> TimedFunction[P, T]:
        return TimedFunction(func, timing=attrs.evolve(self, **kwargs))

    @__call__.register(Iterable)
    def _[T](
        self, iterable: Iterable[T], /, total: int | None = None, **kwargs
    ) -> TimedIterable[T]:
        return TimedIterable(iterable, timing=attrs.evolve(self, **kwargs), total=total)



import math
import statistics
from collections.abc import Iterator, Mapping, Sequence
from typing import ClassVar

import attrs

from liblaf.grapes import human


@attrs.define
class Statistics(Mapping[str, str]):
    STATISTICS: ClassVar[Sequence[str]] = (
        "max",
        "mean",
        "mean_std",
        "median",
        "min",
        "stdev",
        "sum",
    )
    data: Sequence[float] = attrs.field(factory=list)

    def __getitem__(self, key: str, /) -> str:
        try:
            value: float | str | None = getattr(self, key, None)
        except statistics.StatisticsError:
            return human.human_duration(math.nan)
        if value is None:
            raise KeyError(key)
        if isinstance(value, (int, float)):
            value = human.human_duration(value)
        return value

    def __iter__(self) -> Iterator[str]:
        yield from self.STATISTICS

    def __len__(self) -> int:
        return len(self.STATISTICS)

    def pretty_name(self, key: str) -> str:
        if key == "mean_std":
            return "mean"
        return key

    @property
    def max(self) -> float:
        return max(self.data)

    @property
    def mean(self) -> float:
        return statistics.mean(self.data)

    @property
    def mean_std(self) -> str:
        if len(self.data) == 0:
            return human.human_duration(math.nan)
        if len(self.data) == 1:
            return human.human_duration(self.data[0])
        human_mean: str = human.human_duration(self.mean)
        human_stdev: str = human.human_duration(self.stdev)
        return f"{human_mean} ± {human_stdev}"

    @property
    def median(self) -> float:
        return statistics.median(self.data)

    @property
    def min(self) -> float:
        return min(self.data)

    @property
    def stdev(self) -> float:
        return statistics.stdev(self.data)

    @property
    def sum(self) -> float:
        return sum(self.data)



from collections.abc import Iterable
from typing import Literal

from liblaf.grapes import timing
from liblaf.grapes.logging import depth_tracker

from ._progress import Progress


@depth_tracker
def track[T](
    sequence: Iterable[T],
    total: float | None = None,
    completed: int = 0,
    description: str = "Working...",
    update_period: float = 0.1,
    *,
    progress: Progress | None = None,
    timer: timing.Timer | Literal[False] | None = None,
) -> Iterable[T]:
    if timer is None:
        timer = timing.timer(label=description)
    if progress is None:
        progress = Progress(timer=timer)
    with progress, depth_tracker():
        yield from progress.track(
            sequence,
            total=total,
            completed=completed,
            description=description,
            update_period=update_period,
            timer=timer,
        )



from typing import Unpack

import loguru

from liblaf.grapes.logging.filters import make_filter


def file_handler(
    **kwargs: Unpack["loguru.FileHandlerConfig"],
) -> "loguru.FileHandlerConfig":
    kwargs["filter"] = make_filter(kwargs.get("filter"))
    kwargs.setdefault("mode", "w")
    return kwargs



import datetime
import functools
from collections.abc import Callable
from typing import Any, Protocol, TypedDict, Unpack, overload

import joblib

from liblaf.grapes.conf import config


class MemorizedFunc[**P, T](Protocol):
    @property
    def memory(self) -> joblib.Memory: ...
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T: ...


class CacheKwargs(TypedDict, total=False):
    pass


class ReduceSizeKwargs(TypedDict, total=False):
    bytes_limit: int | str | None
    items_limit: int | None
    age_limit: datetime.timedelta | None


@overload
def cache[**P, T](
    *,
    memory: joblib.Memory | None = None,
    reduce_size: ReduceSizeKwargs | None = None,
    **kwargs: Unpack[CacheKwargs],
) -> Callable[[Callable[P, T]], MemorizedFunc[P, T]]: ...
@overload
def cache[**P, T](
    func: Callable[P, T],
    /,
    *,
    memory: joblib.Memory | None = None,
    reduce_size: ReduceSizeKwargs | None = None,
    **kwargs: Unpack[CacheKwargs],
) -> MemorizedFunc[P, T]: ...
def cache[**P, T](
    func: Callable[P, T] | None = None,
    /,
    *,
    memory: joblib.Memory | None = None,
    reduce_size: ReduceSizeKwargs | None = None,
    **kwargs: Unpack[CacheKwargs],
) -> Any:
    if func is None:
        return functools.partial(
            cache, memory=memory, reduce_size=reduce_size, **kwargs
        )
    if memory is None:
        memory = joblib.Memory(config.joblib_memory_location)
    if reduce_size is None:
        reduce_size = {"bytes_limit": config.joblib_memory_bytes_limit}

    @memory.cache(**kwargs)
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        ret: T = func(*args, **kwargs)
        memory.reduce_size(**reduce_size)
        return ret

    wrapper.memory = memory  # pyright: ignore[reportAttributeAccessIssue]
    return wrapper



from ._collection import SizedIterable
from ._decorator import (
    Decorator,
    DecoratorWithArguments,
    DecoratorWithOptionalArguments,
)
from ._misc import ClassInfo, PathLike

__all__ = [
    "ClassInfo",
    "Decorator",
    "DecoratorWithArguments",
    "DecoratorWithOptionalArguments",
    "PathLike",
    "SizedIterable",
]



from ._as_iterable import as_iterable
from ._as_sequence import as_sequence
from ._first_not_none import first_not_none
from ._generator_to_list import generator_to_list
from ._iterable_wrapper import IterableWrapper

__all__ = [
    "IterableWrapper",
    "as_iterable",
    "as_sequence",
    "first_not_none",
    "generator_to_list",
]



import enum
import sys
from pathlib import Path

import platformdirs
import pydantic
import pydantic_settings as ps


def joblib_memory_location() -> Path:
    return platformdirs.user_cache_path(appname="joblib")


def log_file() -> Path:
    entrypoint: Path = Path(sys.argv[0])
    return entrypoint.parent / "run.log"


class LogLevel(enum.StrEnum):
    TRACE = "TRACE"
    DEBUG = "DEBUG"
    INFO = "INFO"
    SUCCESS = "SUCCESS"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class Config(ps.BaseSettings):
    model_config = ps.SettingsConfigDict(env_prefix="PLAYGROUND_")

    joblib_memory_bytes_limit: int | str | None = pydantic.Field(default="1G")

    joblib_memory_location: Path = pydantic.Field(
        default_factory=joblib_memory_location
    )

    log_file: Path | None = pydantic.Field(default=None)

    log_level: LogLevel = pydantic.Field(default=LogLevel.INFO)


config = Config()



import re
import sys
from pathlib import Path

from environs import env
from typing_extensions import deprecated

from liblaf.grapes.typed import PathLike


@deprecated("Use `liblaf-cherries` instead.")
def project_root(
    start: Path | None = None, name: str | re.Pattern = r"\.git|src|playground"
) -> Path:
    if (root := env.path("PROJECT_ROOT", default=None)) and root.exists():
        return root
    if start is None:
        start = Path(sys.argv[0])
    start = start.absolute()
    path: Path = Path(start)
    if isinstance(name, str):
        name = re.compile(name)
    while name.match(path.name) is None:
        if path.parent == path:  # not found
            return start.parent
        path = path.parent
    return path.parent


@deprecated("Use `liblaf-cherries` instead.")
def resolve_project_path(
    relative: PathLike = ".",
    *,
    start: Path | None = None,
    name: str | re.Pattern = r"\.git|src|playground",
) -> Path:
    root: Path = project_root(start=start, name=name)
    return root / relative



import functools
from collections.abc import Callable, Mapping
from typing import Any, NoReturn, overload

import attrs
import sortedcontainers
from rich.text import Text

from liblaf.grapes import pretty
from liblaf.grapes.typed import Decorator


@attrs.frozen
class Function:
    condition: Callable[..., bool]
    function: Callable
    precedence: int = 0


class NotFoundLookupError(LookupError):
    func: Callable
    args: tuple
    kwargs: Mapping

    def __init__(self, func: Callable, args: tuple, kwargs: Mapping) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __str__(self) -> str:
        return f"{self.pretty_call.plain} could not be resolved."

    @property
    def pretty_call(self) -> Text:
        return pretty.call(self.func, self.args, self.kwargs)


def _fallback(func: Callable) -> Callable[..., NoReturn]:
    def fallback(*args, **kwargs) -> NoReturn:
        raise NotFoundLookupError(func, args, kwargs)

    return fallback


class ConditionalDispatcher:
    fallback: Callable
    functions: sortedcontainers.SortedList[Function]

    def __init__(self) -> None:
        self.functions = sortedcontainers.SortedList(key=lambda f: -f.precedence)

    def __call__(self, *args, **kwargs) -> Any:
        for func in self.functions:
            try:
                if func.condition(*args, **kwargs):
                    return func.function(*args, **kwargs)
            except TypeError:
                continue
        return self.fallback(*args, **kwargs)

    @overload
    def final[**P, T](
        self, fn: Callable[P, T], /, *, fallback: bool = False
    ) -> Callable[P, T]: ...
    @overload
    def final(self, /, *, fallback: bool = False) -> Decorator: ...
    def final[**P, T](
        self, fn: Callable[P, T] | None = None, /, *, fallback: bool = False
    ) -> Callable:
        if fn is None:
            return functools.partial(self.final, fallback=fallback)
        if fallback:
            self.fallback = fn
        else:
            self.fallback = _fallback(fn)
        functools.update_wrapper(self, fn)
        return self

    def register(
        self, condition: Callable[..., bool], *, precedence: int = 0
    ) -> Decorator:
        def decorator[**P, T](fn: Callable[P, T]) -> Callable[P, T]:
            self.functions.add(Function(condition, fn, precedence))
            return fn

        return decorator



from typing import Any

import attrs


@attrs.define
class MatchError(ValueError):
    value: Any
    cls: str | type = "match"

    def __str__(self) -> str:
        cls: str = self.cls if isinstance(self.cls, str) else self.cls.__qualname__
        return f"{self.value!r} is not a valid {cls}"



from collections.abc import Iterable
from typing import overload

import loguru

from ._base import BaseTimer, Callback

@overload
def log_record(
    *,
    idx: int = -1,
    level: int | str = "DEBUG",
    logger: loguru.Logger | None = None,
    threshold: float = 0.1,  # seconds
) -> Callback: ...
@overload
def log_record(
    records: BaseTimer,
    /,
    *,
    idx: int = -1,
    level: int | str = "DEBUG",
    logger: loguru.Logger | None = None,
    threshold: float = 0.1,  # seconds
) -> None: ...
@overload
def log_summary(
    *,
    level: int | str = "INFO",
    logger: loguru.Logger | None = None,
    stats: Iterable[str] = ...,
) -> Callback: ...
@overload
def log_summary(
    records: BaseTimer,
    /,
    *,
    level: int | str = "INFO",
    logger: loguru.Logger | None = None,
    stats: Iterable[str] = ...,
) -> None: ...

__all__ = ["log_record", "log_summary"]



def first_not_none[T](*args: T | None) -> T:
    """...

    References:
        1. [`more_itertools.first_true`](https://more-itertools.readthedocs.io/en/stable/api.html#more_itertools.first_true)
    """
    return next(arg for arg in args if arg is not None)



import contextlib
import enum
from collections.abc import Sequence
from typing import override

import loguru
from loguru import logger


class LogLevel(enum.StrEnum):
    @override
    @staticmethod
    def _generate_next_value_(
        name: str, start: int, count: int, last_values: list[str]
    ) -> str:
        return name.upper()

    # ref: <https://github.com/Delgan/loguru/blob/master/loguru/_defaults.py>
    TRACE = enum.auto()
    DEBUG = enum.auto()
    INFO = enum.auto()
    SUCCESS = enum.auto()
    WARNING = enum.auto()
    ERROR = enum.auto()
    CRITICAL = enum.auto()


DEFAULT_LEVELS: Sequence["loguru.LevelConfig"] = [
    {"name": "ICECREAM", "no": 15, "color": "<magenta><bold>", "icon": "🍦"}
]


def add_level(
    name: str, no: int, color: str | None = None, icon: str | None = None
) -> None:
    """Add a new logging level to the logger.

    Args:
        name: The name of the new logging level.
        no: The numeric value of the new logging level.
        color: The color associated with the new logging level.
        icon: The icon associated with the new logging level.
    """
    with contextlib.suppress(ValueError):
        logger.level(name, no, color=color, icon=icon)



import functools
import logging
import sys
import types
from collections.abc import Sequence
from typing import Protocol, Unpack

import cytoolz as toolz
import loguru
from loguru import logger

from liblaf.grapes.conf import config

from ._intercept import setup_loguru_logging_intercept
from ._level import DEFAULT_LEVELS, add_level
from ._std import clear_handlers
from .handler import file_handler, rich_handler


def init_loguru(
    handlers: Sequence["loguru.HandlerConfig"] | None = None,
    levels: Sequence["loguru.LevelConfig"] | None = None,
    *,
    enable_link: bool = True,
    **kwargs: Unpack["loguru.BasicHandlerConfig"],
) -> None:
    traceback_install()
    if handlers is None:
        handlers: list[loguru.HandlerConfig] = [
            rich_handler(enable_link=enable_link, **kwargs)
        ]
        if config.log_file:
            handlers.append(
                file_handler(**toolz.merge(kwargs, {"sink": config.log_file}))
            )
    logger.configure(handlers=handlers)
    for lvl in levels or DEFAULT_LEVELS:
        add_level(**lvl)
    setup_loguru_logging_intercept(kwargs.get("level", logging.NOTSET))
    clear_handlers()


def traceback_install(
    except_level: int | str | None = "CRITICAL",
    except_message: str = "",
    *,
    unraisable_level: int | str | None = "ERROR",
) -> None:
    if except_level is not None:
        sys.excepthook = functools.partial(
            excepthook, level=except_level, message=except_message
        )
    if unraisable_level is not None:
        sys.unraisablehook = functools.partial(unraisablehook, level=unraisable_level)


def excepthook(
    exc_type: type[BaseException],
    exc_value: BaseException,
    traceback: types.TracebackType,
    *,
    level: int | str = "CRITICAL",
    message: str = "",
) -> None:
    logger.opt(exception=(exc_type, exc_value, traceback)).log(level, message)


class UnraisableHookArgs(Protocol):
    exc_type: type[BaseException]
    exc_value: BaseException | None
    exc_traceback: types.TracebackType | None
    err_msg: str | None
    object: object


def unraisablehook(
    unraisable: UnraisableHookArgs, /, *, level: int | str = "ERROR"
) -> None:
    if logger is None:
        # workaround for "Error ignored in: ..."
        return
    logger.opt(
        exception=(unraisable.exc_type, unraisable.exc_value, unraisable.exc_traceback)
    ).log(
        level,
        "{err_msg}: {object!r}",
        err_msg=unraisable.err_msg or "Exception ignored in",
        object=unraisable.object,
    )



from . import filters, handler, sink
from ._depth_tracker import depth_tracker
from ._icecream import init_icecream
from ._init import init_logging
from ._init_loguru import init_loguru, traceback_install
from ._intercept import InterceptHandler, setup_loguru_logging_intercept
from ._level import LogLevel, add_level
from ._std import clear_handlers
from .filters import (
    Filter,
    as_filter_func,
    filter_all,
    filter_any,
    filter_once,
    make_filter,
)
from .handler import file_handler, jsonl_handler, rich_handler
from .sink import (
    LevelColumn,
    LocationColumn,
    LoguruRichHandler,
    MessageColumn,
    RichLoggingColumn,
    TimeColumn,
)

__all__ = [
    "Filter",
    "InterceptHandler",
    "LevelColumn",
    "LocationColumn",
    "LogLevel",
    "LoguruRichHandler",
    "MessageColumn",
    "RichLoggingColumn",
    "TimeColumn",
    "add_level",
    "as_filter_func",
    "clear_handlers",
    "depth_tracker",
    "file_handler",
    "filter_all",
    "filter_any",
    "filter_once",
    "filters",
    "handler",
    "init_icecream",
    "init_logging",
    "init_loguru",
    "jsonl_handler",
    "make_filter",
    "rich_handler",
    "setup_loguru_logging_intercept",
    "sink",
    "traceback_install",
]



import inspect
import itertools
import logging
from collections.abc import Iterable

import loguru._logger
from loguru import logger


class InterceptHandler(logging.Handler):
    """A logging handler that intercepts log messages and redirects them to Loguru.

    This handler is designed to be compatible with the standard logging framework and allows the use of Loguru for logging while maintaining compatibility with existing logging configurations.

    References:
        1. [Overview — loguru documentation](https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging)
    """

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a log record.

        This function is called by the logging framework to handle a log record.
        It maps the standard logging levels to Loguru levels and finds the caller
        frame from where the log message originated. Finally, it logs the message
        using Loguru.

        Args:
            record: The log record to be emitted.
        """
        if logger is None:
            # workaround for "Error ignored in: ..."
            return

        # Get corresponding Loguru level if it exists.
        level: str | int
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


def setup_loguru_logging_intercept(
    level: int | str = logging.NOTSET, modules: Iterable[str] = ()
) -> None:
    """Sets up logging interception using Loguru.

    This function configures the logging module to use Loguru for handling log messages. It sets the logging level and replaces the handlers for the specified modules with an InterceptHandler that redirects log messages to Loguru.

    Args:
        level: The logging level to set.
        modules: A list of module names whose loggers should be intercepted.

    References:
        1. [loguru-logging-intercept/loguru_logging_intercept.py at f358b75ef4162ea903bf7a3298c22b1be83110da · MatthewScholefield/loguru-logging-intercept](https://github.com/MatthewScholefield/loguru-logging-intercept/blob/f358b75ef4162ea903bf7a3298c22b1be83110da/loguru_logging_intercept.py#L35C5-L42)
    """
    core: loguru._logger.Core = logger._core  # pyright: ignore[reportAttributeAccessIssue] # noqa: SLF001
    for lvl in core.levels.values():
        logging.addLevelName(lvl.no, lvl.name)
    logging.basicConfig(level=level, handlers=[InterceptHandler()])
    for logger_name in itertools.chain(("",), modules):
        mod_logger: logging.Logger = logging.getLogger(logger_name)
        mod_logger.handlers = [InterceptHandler(level=level)]
        mod_logger.propagate = False



from collections.abc import Mapping

import loguru

from ._as_filter_func import as_filter_func
from ._composite import filter_all
from ._once import filter_once
from .typed import Filter

DEFAULT_FILTER: "loguru.FilterDict" = {
    "": "INFO",
    "__main__": "TRACE",
    "liblaf": "DEBUG",
}


def make_filter(
    filter_: Filter | None = None, *, inherit: bool = True
) -> "loguru.FilterFunction | None":
    if inherit:
        if filter_ is None:
            filter_ = DEFAULT_FILTER
        elif isinstance(filter_, Mapping):
            filter_ = {**DEFAULT_FILTER, **filter_}
    filter_ = as_filter_func(filter_)  # pyright: ignore[reportArgumentType]
    if inherit:
        filter_ = filter_all(filter_, filter_once())
    return filter_



import functools
from pathlib import Path

from environs import Env


@functools.lru_cache
def init_env(path: str | Path | None = ".env", prefix: str | None = None) -> Env:
    env = Env(prefix=prefix)
    env.read_env(path)
    return env



import contextlib
from collections.abc import Generator
from typing import Any

import etils.epy


@contextlib.contextmanager
def optional_imports(
    name: str = "liblaf-grapes", extra: str | None = None
) -> Generator[None, Any, None]:
    """Context manager to handle optional imports gracefully.

    This context manager allows you to attempt to import optional dependencies
    and handle `ImportError` exceptions in a user-friendly manner. If an import
    fails, it provides a helpful message indicating the missing dependency and
    suggests how to install it.

    Args:
        name: The name of the package that contains the optional dependency.
        extra: An optional string specifying the extra requirements to install the missing dependency.

    Raises:
        ImportError: If an optional dependency is missing, an `ImportError` is raised with a helpful message suggesting how to install it.
    """
    try:
        yield
    except ImportError as err:
        suffix: str = f"Missing optional dependency `{err.name}`."
        if extra is not None:
            suffix += (
                f"\nMake sure to install `{name}` using `pip install {name}[{extra}]`."
            )
        etils.epy.reraise(err, suffix=suffix)



from typing import Any, TypedDict, Unpack

import attrs
import cytoolz as toolz
import wadler_lindig as wl


class PdocKwargs(TypedDict, total=True):
    """.

    References:
        1. <https://docs.kidger.site/wadler_lindig/api/#wadler_lindig.pdoc>
    """

    indent: int
    hide_defaults: bool
    show_type_module: bool
    show_dataclass_module: bool


@attrs.frozen
class WithRepr:
    string: str = attrs.field()

    def __repr__(self) -> str:
        return self.string


UNINITIALIZED = WithRepr("<uninitialized>")


def pdoc_attrs(self: Any, **kwargs: Unpack[PdocKwargs]) -> wl.AbstractDoc:
    """.

    References:
        1. <https://github.com/patrick-kidger/wadler_lindig/blob/0226340d56f0c18e10cd4d375cf7ea25818359b8/wadler_lindig/_definitions.py#L308-L326>
    """
    cls: type = type(self)
    objs: list[tuple[str, Any]] = []
    for field in attrs.fields(cls):
        field: attrs.Attribute
        if not field.repr:
            continue
        value: Any = getattr(self, field.name, UNINITIALIZED)
        if kwargs["hide_defaults"] and value is field.default:
            continue
        objs.append((field.name, value))
    name_kwargs: PdocKwargs = toolz.assoc(
        kwargs, "show_type_module", kwargs["show_dataclass_module"]
    )  # pyright: ignore[reportAssignmentType]
    return wl.bracketed(
        begin=wl.pdoc(cls, **name_kwargs) + wl.TextDoc("("),
        docs=wl.named_objs(objs, **kwargs),
        sep=wl.comma,
        end=wl.TextDoc(")"),
        indent=kwargs["indent"],
    )


class WadlerLindigMixin:
    def __repr__(self) -> str:
        return wl.pformat(self)

    def __pdoc__(self, **kwargs) -> wl.AbstractDoc:
        return pdoc_attrs(self, **kwargs)



import os
from pathlib import Path
from typing import Any

from typing_extensions import TypeIs

from liblaf.grapes.typed import PathLike


def as_path(path: PathLike, *, expend_user: bool = True) -> Path:
    path = Path(path)
    if expend_user:
        path = path.expanduser()
    return path


def is_path_like(obj: Any) -> TypeIs[PathLike]:
    return isinstance(obj, (str, os.PathLike))



from ._clone_signature import clone_signature
from ._conditional_dispatcher import ConditionalDispatcher
from ._memory import MemorizedFunc, cache

__all__ = ["ConditionalDispatcher", "MemorizedFunc", "cache", "clone_signature"]



from ._abc import Serde
from ._json import JSON, json
from ._pydantic import load_pydantic, save_pydantic
from ._serde import Auto, auto, load, save
from ._toml import TOML, toml
from ._yaml import YAML, yaml

__all__ = [
    "JSON",
    "TOML",
    "YAML",
    "Auto",
    "Serde",
    "auto",
    "json",
    "load",
    "load_pydantic",
    "save",
    "save_pydantic",
    "save_pydantic",
    "toml",
    "yaml",
]



import datetime
import types
import unittest.mock
from collections.abc import Generator, Sequence
from typing import Protocol, override

import attrs
import loguru
import wadler_lindig as wl
from rich.console import Console, RenderableType
from rich.highlighter import Highlighter, ReprHighlighter
from rich.text import Text
from rich.traceback import Traceback

from liblaf.grapes import pretty


class RichLoggingColumn(Protocol):
    def render(self, record: "loguru.Record") -> RenderableType: ...


@attrs.define
class LoguruRichHandler:
    console: Console = attrs.field(factory=lambda: pretty.get_console(stderr=True))
    columns: Sequence[RichLoggingColumn] = attrs.field(
        factory=lambda: [TimeColumn(), LevelColumn(), LocationColumn(), MessageColumn()]
    )

    def __call__(self, message: "loguru.Message") -> None:
        record: loguru.Record = message.record
        # TODO: `console.print()` is slow
        self.console.print(
            *self.render(record), overflow="ignore", no_wrap=True, crop=False
        )
        if (excpetion := self.render_exception(record)) is not None:
            self.console.print(excpetion)

    def render(self, record: "loguru.Record") -> Generator[RenderableType]:
        for column in self.columns:
            yield column.render(record)

    def render_exception(self, record: "loguru.Record") -> RenderableType | None:
        exception: loguru.RecordException | None = record["exception"]
        if exception is None:
            return None
        exc_type: type[BaseException] | None
        exc_value: BaseException | None
        traceback: types.TracebackType | None
        exc_type, exc_value, traceback = exception
        if exc_type is None or exc_value is None:
            return None

        # ? dirty hack to avoid long `repr()` output
        # ref: <https://github.com/Textualize/rich/discussions/3774>
        with unittest.mock.patch("rich.pretty.repr", new=wl.pformat):
            rich_tb: Traceback = Traceback.from_exception(
                exc_type=exc_type,
                exc_value=exc_value,
                traceback=traceback,
                width=self.console.width,
                code_width=self.console.width,
                show_locals=True,
            )

        # ? dirty hack to support ANSI in exception messages
        for stack in rich_tb.trace.stacks:
            if pretty.has_ansi(stack.exc_value):
                stack.exc_value = Text.from_ansi(stack.exc_value)  # pyright: ignore[reportAttributeAccessIssue]
        return rich_tb


class TimeColumn(RichLoggingColumn):
    @override
    def render(self, record: "loguru.Record") -> RenderableType:
        elapsed: datetime.timedelta = record["elapsed"]
        hh: int
        mm: int
        ss: int
        mm, ss = divmod(int(elapsed.total_seconds()), 60)
        hh, mm = divmod(mm, 60)
        return Text(
            f"{hh:02d}:{mm:02d}:{ss:02d}.{elapsed.microseconds:06d}", style="log.time"
        )


@attrs.define
class LevelColumn(RichLoggingColumn):
    @override
    def render(self, record: "loguru.Record") -> RenderableType:
        level: str = record["level"].name
        return Text(f"{level:<8}", style=f"logging.level.{level.lower()}")


@attrs.define
class LocationColumn(RichLoggingColumn):
    enable_link: bool = attrs.field(default=True)

    @override
    def render(self, record: "loguru.Record") -> RenderableType:
        location: Text = pretty.location(
            name=record["name"],
            function=record["function"],
            line=record["line"],
            file=record["file"].path,
            enable_link=self.enable_link,
        )
        location.style = "log.path"
        return location


@attrs.define
class MessageColumn(RichLoggingColumn):
    highlighter: Highlighter = attrs.field(factory=ReprHighlighter)

    @override
    def render(self, record: "loguru.Record") -> RenderableType:
        if (rich := record["extra"].get("rich")) is not None:
            return rich
        message: RenderableType = record["message"].strip()
        if "\x1b" in message:
            return Text.from_ansi(message)
        if record["extra"].get("markup", True):
            message = Text.from_markup(message)
        if highlighter := record["extra"].get("highlighter", self.highlighter):
            message = highlighter(message)
        return message



import contextlib
import types
from collections.abc import Callable, Iterable
from typing import Self, overload, override

from ._base import BaseTimer
from ._function import TimedFunction
from ._iterable import TimedIterable

class Timer(contextlib.AbstractContextManager, BaseTimer):
    @override
    def __enter__(self) -> Self: ...
    @override
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
        /,
    ) -> None: ...
    @overload
    def __call__[**P, T](
        self, func: Callable[P, T], /, **kwargs
    ) -> TimedFunction[P, T]: ...
    @overload
    def __call__[T](
        self, iterable: Iterable[T], /, total: int | None = None, **kwargs
    ) -> TimedIterable[T]: ...



import collections
from collections.abc import Callable, Generator, Iterable, Mapping, Sequence

import attrs
import loguru

from liblaf.grapes import const, human
from liblaf.grapes.logging import depth_tracker

from . import callback
from ._get_time import TimerName, get_time
from ._statistics import Statistics
from .defaults import DEFAULT_STATS, DEFAULT_TIMERS

type Callback = Callable[[BaseTimer], None] | const.NopType


@attrs.define
class BaseTimer:
    label: str | None = attrs.field(default=None)
    timers: Sequence[TimerName] = attrs.field(default=DEFAULT_TIMERS, kw_only=True)
    cb_start: Callback | None = attrs.field(default=None, kw_only=True)
    cb_stop: Callback | None = attrs.field(
        default=callback.log_record,
        converter=attrs.converters.default_if_none(callback.log_record),
        kw_only=True,
    )
    cb_finish: Callback | None = attrs.field(
        default=callback.log_summary,
        converter=attrs.converters.default_if_none(callback.log_summary),
        kw_only=True,
    )
    _records: collections.defaultdict[TimerName, list[float]] = attrs.field(
        init=False, factory=lambda: collections.defaultdict(list)
    )
    _time_start: dict[TimerName, float] = attrs.field(init=False, factory=dict)
    _time_stop: dict[TimerName, float] = attrs.field(init=False, factory=dict)

    @property
    def default_timer(self) -> TimerName:
        return self.timers[0]

    # region Timer

    def elapsed(self, timer: TimerName | None = None) -> float:
        timer = timer or self.default_timer
        time_stop: float
        if timer in self._time_stop:
            time_stop = self._time_stop[timer]
        else:
            time_stop = get_time(timer)
        return time_stop - self._time_start[timer]

    @depth_tracker
    def start(self) -> None:
        for timer in self.timers:
            self._time_start[timer] = get_time(timer)
        if callable(self.cb_start):
            self.cb_start(self)

    @depth_tracker
    def stop(self) -> None:
        for timer in self.timers:
            self._time_stop[timer] = get_time(timer)
            elapsed: float = self._time_stop[timer] - self._time_start[timer]
            self._records[timer].append(elapsed)
        if callable(self.cb_stop):
            self.cb_stop(self)

    @depth_tracker
    def finish(self) -> None:
        if callable(self.cb_finish):
            self.cb_finish(self)

    # endregion Timer

    # region Container

    @property
    def height(self) -> int:
        return len(self.column())

    @property
    def width(self) -> int:
        return len(self.timers)

    def clear(self) -> None:
        self._records.clear()
        self._time_start.clear()
        self._time_stop.clear()

    def column(self, timer: TimerName | None = None) -> Sequence[float]:
        timer = timer or self.default_timer
        return self._records[timer]

    def iter_columns(self) -> Generator[tuple[TimerName, Sequence[float]]]:
        for timer in self.timers:
            yield timer, self.column(timer)

    def iter_rows(self) -> Generator[Mapping[TimerName, float]]:
        for idx in range(self.height):
            yield self.row(idx)

    def row(self, idx: int = -1) -> Mapping[TimerName, float]:
        return {k: v[idx] for k, v in self._records.items()}

    # endregion Container

    # region Human-Readable

    def human_record(self, idx: int = -1) -> TimerName:
        label: TimerName = self.label or "Timer"
        items: list[TimerName] = []
        for k, v in self.row(idx).items():
            hd: TimerName = human.human_duration(v)
            items.append(f"{k}: {hd}")
        return f"{label} > {', '.join(items)}"

    def human_summary(self, stats: Iterable[str] = DEFAULT_STATS) -> str:
        label: TimerName = self.label or "Timer"
        header: TimerName = f"{label} (count: {self.height})"
        if self.height == 0:
            return header
        lines: list[TimerName] = []
        for timer in self.timers:
            stats_str: list[str] = []
            for stat in stats:
                stats_mapping: Statistics = self.stats(timer)
                name: str = stats_mapping.pretty_name(stat)
                value: str = stats_mapping[stat]
                stats_str.append(f"{name}: {value}")
            line: str = f"{timer} > {', '.join(stats_str)}"
            lines.append(line)
        if self.width == 1:
            return f"{header} {lines[0]}"
        return f"{header}\n" + "\n".join(lines)

    @depth_tracker
    def log_record(
        self,
        *,
        idx: int = -1,
        level: int | str = "DEBUG",
        logger: "loguru.Logger | None" = None,
        threshold: float = 0.1,  # seconds
    ) -> None:
        if self.elapsed() < threshold:
            return
        if logger is None:
            logger = loguru.logger.opt(depth=depth_tracker.depth)
        logger.log(level, self.human_record(idx))

    @depth_tracker
    def log_summary(
        self,
        *,
        level: int | str = "INFO",
        logger: "loguru.Logger | None" = None,
        stats: Iterable[str] = DEFAULT_STATS,
    ) -> None:
        if logger is None:
            logger = loguru.logger.opt(depth=depth_tracker.depth)
        logger.log(level, self.human_summary(stats))

    # endregion Human-Readable

    # region Statistics

    def stats(self, timer: TimerName | None = None) -> Statistics:
        return Statistics(self.column(timer))

    # endregion Statistics



from . import (
    conf,
    const,
    environ,
    error,
    functools,
    git,
    human,
    itertools,
    logging,
    path,
    pretty,
    serde,
    text,
    timing,
    tqdm,
    typed,
)
from ._version import __version__, __version_tuple__, version, version_tuple
from .conf import Config, config
from .const import MISSING, NOP, MissingType, NopType, Sentinel, nop
from .deps import has_module, optional_imports
from .environ import init_env
from .error import MatchError
from .functools import ConditionalDispatcher, MemorizedFunc, cache, clone_signature
from .human import (
    human_count,
    human_duration,
    human_duration_series,
    human_duration_unit_precision,
    human_duration_with_variance,
    human_throughout,
)
from .itertools import (
    IterableWrapper,
    as_iterable,
    as_sequence,
    first_not_none,
    generator_to_list,
)
from .logging import LogLevel, file_handler, init_logging, jsonl_handler, rich_handler
from .path import as_path, is_path_like, project_root, resolve_project_path
from .pretty import WadlerLindigMixin, caller_location, get_console, pdoc_attrs
from .serde import json, load, load_pydantic, save, save_pydantic, toml, yaml
from .text import strip_comments
from .timing import (
    BaseTimer,
    TimedFunction,
    TimedIterable,
    Timer,
    TimerName,
    get_time,
    timer,
)
from .tqdm import Progress, RateColumn, len_safe, parallel, track
from .typed import (
    ClassInfo,
    Decorator,
    DecoratorWithArguments,
    DecoratorWithOptionalArguments,
    PathLike,
    SizedIterable,
)

__all__ = [
    "MISSING",
    "NOP",
    "BaseTimer",
    "ClassInfo",
    "ConditionalDispatcher",
    "Config",
    "Decorator",
    "DecoratorWithArguments",
    "DecoratorWithOptionalArguments",
    "IterableWrapper",
    "LogLevel",
    "MatchError",
    "MemorizedFunc",
    "MissingType",
    "MissingType",
    "NopType",
    "PathLike",
    "Progress",
    "RateColumn",
    "Sentinel",
    "SizedIterable",
    "TimedFunction",
    "TimedIterable",
    "Timer",
    "TimerName",
    "WadlerLindigMixin",
    "__version__",
    "__version_tuple__",
    "as_iterable",
    "as_path",
    "as_sequence",
    "cache",
    "caller_location",
    "clone_signature",
    "conf",
    "config",
    "const",
    "environ",
    "error",
    "file_handler",
    "first_not_none",
    "functools",
    "generator_to_list",
    "get_console",
    "get_time",
    "git",
    "has_module",
    "human",
    "human_count",
    "human_duration",
    "human_duration_series",
    "human_duration_unit_precision",
    "human_duration_with_variance",
    "human_throughout",
    "init_env",
    "init_logging",
    "is_path_like",
    "itertools",
    "json",
    "jsonl_handler",
    "len_safe",
    "load",
    "load_pydantic",
    "logging",
    "nop",
    "optional_imports",
    "parallel",
    "path",
    "pdoc_attrs",
    "pretty",
    "project_root",
    "resolve_project_path",
    "rich_handler",
    "save",
    "save_pydantic",
    "serde",
    "strip_comments",
    "text",
    "timer",
    "timing",
    "toml",
    "tqdm",
    "track",
    "typed",
    "version",
    "version_tuple",
    "yaml",
]



import enum


class Sentinel(enum.Enum):
    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return f"{type(self).__qualname__}({self.name!r})"


class MissingType(Sentinel):
    MISSING = enum.auto()


MISSING = MissingType.MISSING


class NopType(Sentinel):
    NOP = enum.auto()


NOP = NopType.NOP



import functools
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING

import loguru

from liblaf.grapes.logging import depth_tracker

from .defaults import DEFAULT_STATS

if TYPE_CHECKING:
    from ._base import BaseTimer


@depth_tracker
def log_record(
    records: "BaseTimer | None" = None,
    /,
    *,
    idx: int = -1,
    level: int | str = "DEBUG",
    logger: "loguru.Logger | None" = None,
    threshold: float = 0.1,  # seconds
) -> Callable | None:
    if records is None:
        return functools.partial(
            log_record, idx=idx, level=level, logger=logger, threshold=threshold
        )
    return records.log_record(idx=idx, level=level, logger=logger, threshold=threshold)


@depth_tracker
def log_summary(
    records: "BaseTimer | None" = None,
    /,
    *,
    level: int | str = "INFO",
    logger: "loguru.Logger | None" = None,
    stats: Iterable[str] = DEFAULT_STATS,
) -> Callable | None:
    if records is None:
        return functools.partial(log_summary, level=level, logger=logger, stats=stats)
    return records.log_summary(level=level, logger=logger, stats=stats)


__all__ = ["log_record", "log_summary"]



from . import callback
from ._base import BaseTimer, Callback
from ._function import TimedFunction
from ._get_time import TimerName, get_time
from ._iterable import TimedIterable
from ._main import timer
from ._timer import Timer

__all__ = [
    "BaseTimer",
    "Callback",
    "TimedFunction",
    "TimedIterable",
    "Timer",
    "TimerName",
    "callback",
    "get_time",
    "timer",
    "timer",
]



from collections.abc import Buffer
from pathlib import Path
from typing import Any

from liblaf.grapes.typed import PathLike


class Serde:
    def decode(self, buf: Buffer | str, /, **kwargs) -> Any:
        raise NotImplementedError

    def encode(self, obj: Any, /, **kwargs) -> bytes:
        raise NotImplementedError

    def load(self, path: PathLike, /, **kwargs) -> Any:
        path = Path(path)
        return self.decode(path.read_bytes(), **kwargs)

    def save(self, path: PathLike, obj: Any, /, **kwargs) -> None:
        path = Path(path)
        path.write_bytes(self.encode(obj, **kwargs))



from collections.abc import Callable, Mapping, Sequence

import attrs

from liblaf.grapes import pretty


@attrs.define
class DispatchLookupError(LookupError):
    func: Callable = attrs.field()
    call_args: Sequence = attrs.field(factory=tuple)
    call_kwargs: Mapping = attrs.field(factory=dict)

    def __str__(self) -> str:
        pretty_func_call: str = pretty.call(
            self.func, self.call_args, self.call_kwargs
        ).plain
        return f"`{pretty_func_call}` could not be resolved."



from typing import overload

import loguru

from .typed import Filter

@overload
def as_filter_func(filter_: None, /) -> None: ...
@overload
def as_filter_func(filter_: Filter, /) -> loguru.FilterFunction: ...



from collections.abc import Iterable, Iterator

from liblaf.grapes.logging import depth_tracker

from ._base import BaseTimer


class TimedIterable[T]:
    timing: BaseTimer
    _iterable: Iterable[T]
    _total: int | None = None

    def __init__(
        self, iterable: Iterable[T], /, *, timing: BaseTimer, total: int | None = None
    ) -> None:
        self.timing = timing
        self._iterable = iterable
        self._total = total
        if self.timing.label is None:
            self.timing.label = "Iterable"

    def __contains__(self, x: object, /) -> bool:
        return x in self._iterable  # pyright: ignore[reportOperatorIssue]

    def __len__(self) -> int:
        if self._total is None:
            return len(self._iterable)  # pyright: ignore[reportArgumentType]
        return self._total

    def __iter__(self) -> Iterator[T]:
        with depth_tracker():
            self.timing.start()
            for item in self._iterable:
                yield item
                self.timing.stop()
                self.timing.start()
            self.timing.finish()



import functools
from collections.abc import Callable, Iterable, Sequence
from typing import Any, overload

from ._base import BaseTimer, Callback
from ._function import TimedFunction
from ._iterable import TimedIterable
from ._timer import Timer


@functools.singledispatch
def _timer_dispatch(_: Any, /, *_args, **_kwargs) -> Any:
    raise TypeError


@_timer_dispatch.register(str)
def _(label: str, /, **kwargs) -> Timer:
    return Timer(label=label, **kwargs)


@_timer_dispatch.register(Callable)  # pyright: ignore[reportArgumentType, reportCallIssue]
def _(fn: Callable, /, **kwargs) -> TimedFunction:
    return TimedFunction(fn, timing=BaseTimer(**kwargs))


@_timer_dispatch.register(Iterable)
def _(iterable: Iterable, /, *, total: int | None = None, **kwargs) -> TimedIterable:
    return TimedIterable(iterable, timing=BaseTimer(**kwargs), total=total)


@overload
def timer[**P, T](
    label: str | None = None,
    *,
    timers: Sequence[str] = ("perf",),
    cb_finish: Callback | None = None,
    cb_start: Callback | None = None,
    cb_stop: Callback | None = None,
) -> Timer: ...
@overload
def timer[**P, T](
    fn: Callable[P, T],
    /,
    *,
    label: str | None = None,
    timers: Sequence[str] = ("perf",),
    cb_finish: Callback | None = None,
    cb_start: Callback | None = None,
    cb_stop: Callback | None = None,
) -> TimedFunction[P, T]: ...
@overload
def timer[**P, T](
    iterable: Iterable[T],
    /,
    *,
    label: str | None = None,
    timers: Sequence[str] = ("perf",),
    total: int | None = None,
    cb_finish: Callback | None = None,
    cb_start: Callback | None = None,
    cb_stop: Callback | None = None,
) -> TimedIterable[T]: ...
def timer(*args, **kwargs) -> Any:
    if args:
        return _timer_dispatch(*args, **kwargs)
    return Timer(*args, **kwargs)



from ._ansi import has_ansi
from ._console import get_console
from ._func import call, func
from ._location import caller_location, location
from ._wadler_lindig import WadlerLindigMixin, pdoc_attrs

__all__ = [
    "WadlerLindigMixin",
    "call",
    "caller_location",
    "func",
    "get_console",
    "has_ansi",
    "location",
    "pdoc_attrs",
]
